#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
"""
    Encrypt and decrypt text using a key system as well as a cipher.


    Constants
    ---------
    file_cipher       Name of file with cipher.
    file_pass         Name of the file with e-mail address list.


    Functions
    ---------
    set_up_cipher     Create or change cipher.
    wordEncrypt       Encrypt word into list of keys using cipher.
    wordDecrypt       Decrypts list of keys into original word using cipher.
    sendfail          Send e-mail to e-mail addresses in file_pass, decrypting with file_cipher.


    Example
    -------
    # Setup the encryption and decryption of password
    if not os.path.exists(file_cipher):
        set_up_cipher(file_cipher)

    # Store encrypted password
    if not os.path.exists(file_pass):
        print('')
        prompt = 'Enter password: '
        ipass = getpass.getpass(prompt)
        cpass = wordEncrypt(ipass)
        f = open(pass_file, 'w')
        f.write(' '.join([ str(i) for i in cpass ])+'\n')
        f.close()

    # Read encrypted password
    f = open(file_pass, 'r')
    cpass = f.readline()[:-1]
    f.close()
    password = wordDecrypt([ int(c) for c in cpass.split()])

    # Send fail e-mail
    sendfail('This test', 'Did not work', sender='me@ufz.de')


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014-2015 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  MC, Jun-Dec 2014 - modified from
              http://code.activestate.com/recipes/577954-encrypt-and-decrypt-text-and-text-files-beta/
    Modified, MC, Oct 2015     - sendfail
"""
from .encrypt import file_cipher, file_pass, set_up_cipher, wordEncrypt, wordDecrypt, sendfail

# Information
__author__   = "Matthias Cuntz"
__version__  = '1.1'
__revision__ = "Revision: 2349"
__date__     = 'Date: 08.10.2015'



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

__all__ = ['pca']

def pca(mat, corr=False, ndim=None, rvar=None):
    """
        Principal component analysis (PCA) upon the first dimension of an 2D-array.

        If you have an Nd-array or an array with masked or NaN values,
        you may want to use Andrew Dawson's eofs package:
            https://github.com/ajdawson/eofs


        Definition
        ----------
        def pca(mat, corr=False, ndim=None, rvar=None):


        Input
        -----
        mat       2D-array, where the first dimension is assumed to represent time.


        Optional Input
        --------------
        corr      True: PCA on correlation matrix
                  False: PCA on covariance matrix
        ndim      # of principal components to retain (default: None, i.e. all)
        rvar      Percent (0-1) of cumulative variance to retain (default: None, i.e. > 1, i.e. all)
                  This uses numpy.searchsorted, which searches for lesser than (<) rvar
                  and not lesser equal (<=) rvar.
                  So rvar=1 retains all but the last principal component, and
                  rvar<np.amin(eigenvalues) retains no principal component at all.


        Output
        ------
        principal_components, eigenvalues, eigenvectors

        If the input is nxk array and there is no dimensionality reduction (ndim or rvar), then
        principal_components is nxk, eigenvalues k, eigenvectors kxk arrays.

        If the input is nxk array and dimensionality reduction selects m principal components, then
        principal_components is nxm, eigenvalues k, eigenvectors kxm arrays.


        Examples
        --------
        >>> import numpy as np
        >>> A = np.array([ [2.4,0.7,2.9,2.2,3.0,2.7,1.6,1.1,1.6,0.9],
        ...                [2.5,0.5,2.2,1.9,3.1,2.3,2,1,1.5,1.1] ]).T

        >>> comps, evals, evecs = pca(A)
        >>> Anew = np.dot(comps, evecs.T) + np.mean(A, axis=0)
        >>> print(np.allclose(A, Anew))
        True

        >>> comps, evals, evecs = pca(A, corr=True)
        >>> Anew = np.dot(comps, evecs.T)*np.std(A, axis=0, ddof=1) + np.mean(A, axis=0)
        >>> print(np.allclose(A, Anew))
        True

        >>> comps, evals, evecs = pca(A, corr=True, ndim=1)
        >>> Anew = np.dot(comps, evecs.T)*np.std(A, axis=0, ddof=1) + np.mean(A, axis=0)
        >>> print(np.allclose(A, Anew))
        False
        >>> print(np.allclose(A, Anew, rtol=0.2))
        True

        >>> comps, evals, evecs = pca(A, corr=True, rvar=0.97)
        >>> Anew = np.dot(comps, evecs.T)*np.std(A, axis=0, ddof=1) + np.mean(A, axis=0)
        >>> print(np.allclose(A, Anew, rtol=0.2))
        True


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Nov 2014
    """
    from scipy import linalg
    imat = mat.copy()
    #
    # Check input
    nd = imat.ndim  # save for later
    assert nd > 1, 'array must be at 2D-ND.'
    n, k = imat.shape
    #
    # Covariance or corelation matrix
    n1 = 1./(float(n)-1.)
    if corr:
        imat /= np.std(imat, axis=0, ddof=1)
    imat -= imat.mean(axis=0)
    S  = n1 * np.dot(imat.T, imat)
    #
    # Eigenvectors and eigenvalues with linalg.eigh
    # rather than linlag.eig since S is symmetric.
    # The performance gain is substantial.
    evals, evecs = linalg.eigh(S)
    #
    # Sort eigenvalues in decreasing order
    # Eigenvalues from linalg.eigh are order increasingly
    # while linalg.eig eigenvalues do not have to be sorted.
    # Use argsort to be on the save side.
    idx   = np.argsort(evals)[::-1]
    evals = evals[idx]
    # sort eigenvectors accordingly
    evecs = evecs[:,idx]
    #
    # Select the first m eigenvectors either by number
    # or by percent of explained variance
    if ndim is not None:
        evecs = evecs[:, :ndim]
    if rvar is not None:
        cumevals  = np.cumsum(evals)
        cumevals /= cumevals[-1]
        cut   = np.searchsorted(cumevals, rvar)
        evecs = evecs[:, :cut]
    #
    # Calculate principal components and return
    # (reduced) components, eigenvalues, (reduced) eigenvectors
    return np.dot(evecs.T, imat.T).T, evals, evecs


def check_pca(data, comps, eigenvectors, corr=False, rtol=1e-05, atol=1e-08):
    '''
        Data recovered from PCA against original data.


        Definition
        ----------
        def check_pca(data, comps, eigenvectors, corr=False, rtol=1e-05, atol=1e-08):


        Input
        -----
        def check_pca(data, comps, eigenvectors, corr=False, rtol=1e-05, atol=1e-08):
        data           original 2D data (nxk)
        comps          2D (reduced) principal components (nxm)
        eigenvectors   (Reduced) Eigenvectors (kxm)


        Optional Input
        --------------
        corr      True: PCA on correlation matrix
                  False: PCA on covariance matrix
        rtol      numpy.allclose rtol keyword: relative tolerance.
        atol      numpy.allclose atol keyword: absolute tolerance.


        Output
        ------
        True or False: numpy.allclose(data, data_recovered, rtol=rtol, atol=atol)


        Examples
        --------
        >>> import numpy as np
        >>> A = np.array([ [2.4,0.7,2.9,2.2,3.0,2.7,1.6,1.1,1.6,0.9],
        ...                [2.5,0.5,2.2,1.9,3.1,2.3,2,1,1.5,1.1] ]).T

        >>> comps, evals, evecs = pca(A)
        >>> check_pca(A, comps, evecs)

        >>> comps, evals, evecs = pca(A, corr=True)
        >>> check_pca(A, comps, evecs, corr=True)

        >>> comps, evals, evecs = pca(A, corr=True, ndim=1)
        >>> check_pca(A, comps, evecs, corr=True, rtol=0.2)

        >>> comps, evals, evecs = pca(A, corr=True, rvar=0.97)
        >>> check_pca(A, comps, evecs, corr=True, rtol=0.2)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Nov 2014
    '''
    # reconstruction
    if corr:
        data_check = np.dot(comps, eigenvectors.T)*np.std(data, axis=0, ddof=1) + np.mean(data, axis=0)
    else:
        data_check = np.dot(comps, eigenvectors.T) + np.mean(data, axis=0)
    assert np.allclose(data, data_check, rtol=rtol, atol=atol), 'data != data_check'


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # from scipy import linalg
    # from scipy.misc import lena
    # import matplotlib.pyplot as plt

    # # http://glowingpython.blogspot.sg/2011/07/principal-component-analysis-with-numpy.html
    # A = np.array([ [2.4,0.7,2.9,2.2,3.0,2.7,1.6,1.1,1.6,0.9],
    #                [2.5,0.5,2.2,1.9,3.1,2.3,2,1,1.5,1.1] ]).T

    # fig = plt.figure()

    # comps, evals, evecs = pca(A)
    # check_pca(A, comps, evecs)

    # sub = plt.subplot(221)
    # # every eigenvector describes the direction of a principal component
    # m = np.mean(A, axis=0)
    # sub.plot([0, -evecs[0,0]*1]+m[0], [0, -evecs[0,1]*1]+m[1], 'k--')
    # sub.plot([0,  evecs[1,0]*1]+m[0], [0,  evecs[1,1]*1]+m[1], 'k--')
    # sub.plot(A[:,0], A[:,1], 'bo') # the data
    # sub.axis('equal')
    # # transformed data
    # sub = plt.subplot(222)
    # sub.plot(comps[:,0], comps[:,1], 'ro')
    # sub.axis('equal')

    # comps, evals, evecs = pca(A, corr=True)
    # check_pca(A, comps, evecs, corr=True)

    # sub = plt.subplot(223)
    # # every eigenvector describes the direction of a principal component
    # m = np.mean(A, axis=0)
    # sub.plot([0, -evecs[0,0]*1]+m[0], [0, -evecs[0,1]*1]+m[1], 'k--')
    # sub.plot([0,  evecs[1,0]*1]+m[0], [0,  evecs[1,1]*1]+m[1], 'k--')
    # sub.plot(A[:,0], A[:,1], 'bo') # the data
    # sub.axis('equal')
    # # transformed data
    # sub = plt.subplot(224)
    # sub.plot(comps[:,0], comps[:,1], 'ro')
    # sub.axis('equal')

    
    # # http://glowingpython.blogspot.it/2011/07/pca-and-image-compression-with-numpy.html
    # fig = plt.figure()
    # A = lena().astype(np.float)

    # full_pc = np.size(A, axis=1) # numbers of all the principal components
    # i=1
    # dist = []
    # nn = 50
    # for numpc in range(1,full_pc+nn,nn): # 1, 51, 101 ... full_pc
    #     comps, evals, evecs = pca(A, ndim=numpc)
    #     Ar = np.dot(comps, evecs.T) + np.mean(A, axis=0) # image reconstruction
    #     # Difference in Frobenius norm
    #     dist.append(linalg.norm(A-Ar,'fro'))
    #     # Reconstructed fig with less PCs
    #     ax = plt.subplot(((full_pc+nn)//nn//3)+1, 3, i, frame_on=False)
    #     ax.xaxis.set_major_locator(plt.NullLocator()) # remove ticks
    #     ax.yaxis.set_major_locator(plt.NullLocator())
    #     i += 1
    #     ax.imshow(Ar)
    #     ax.set_title('PCs # '+str(numpc)+'/'+str(full_pc))
    #     plt.gray()

    # check_pca(A, comps, evecs)

    # plt.figure()
    # perc = np.cumsum(evals)/np.sum(evals)
    # dist = np.array(dist)/np.amax(dist)
    # plt.plot(range(perc.size), perc, 'b')
    # plt.plot(range(1,full_pc+50,50), dist, 'r')
    # plt.axis([1,full_pc,0,1.1])

    # plt.figure()
    # plt.imshow(Ar)
    # plt.title('last numpc')
    # plt.gray()

    # plt.figure()
    # plt.imshow(A)
    # plt.title('numpc FULL')
    # plt.gray()

    
    # # other example
    # # http://sebastianraschka.com/Articles/2014_pca_step_by_step.html
    
    # plt.show()



#!/usr/bin/env python
import numpy as np
from pyjams.const import T25


__all__ = ['dielectric_water']


def dielectric_water(T):
    """
        Dielectric constant of liquid water in F m^-1.


        Definition
        ----------
        def dielectric_water(T):


        Input
        -----
        list/ND-array of temperatures [K]


        Output
        ------
        list/ND-array of dielectric constants of liquid water [F m^-1]


        Restrictions
        ------------
        Valid between 0 degC and 60 degC


        References
        ----------
        Kaatze, U (2007) Reference liquids for the calibration of dielectric
                         sensors and measurement instruments,
                         Measurement Science and Technology, 18


        Examples
        --------
        >>> from autostring import astr
        >>> print(astr(dielectric_water(300.), 3))
        77.693


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014-2015 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Mar 2014 - in const
                  MC, Mar 2015 - own subroutine
    """
    if isinstance(T, np.ma.masked_array):
        return 78.35*np.ma.exp(-4.55e-3*(T-T25))
    else:
        return 78.35*np.exp(-4.55e-3*(T-T25))


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
Divide two arrays, return `otherwise` if division by 0.

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2012-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   division
   div

History
    * Written Jan 2012 by Matthias Cuntz (mc (at) macu (dot) de)
    * Added wrapper div, May 2012, Matthias Cuntz
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Do not return masked array if no masked array given,
      Oct 2014, Matthias Cuntz
    * Added two-digit year, Nov 2018, Matthias Cuntz
    * Removed bug that non-masked array was returned if masked array given,
      Sep 2015, Matthias Cuntz
    * Using numpy docstring format, May 2020, Matthias Cuntz
    * Output type is input type, Oct 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz
    * Use helper functions input2array and array2input,
      Mar 2022, Matthias Cuntz
    * Do not use geterr/seterr, Mar 2022, Matthias Cuntz

"""
import numpy as np
from .helper import input2array, array2input
# from helper import input2array, array2input


__all__ = ['division', 'div']


def division(a, b, otherwise=np.nan, prec=0.):
    """
    Divide two arrays, return `otherwise` if division by 0.

    Parameters
    ----------
    a : array_like
        Enumerator
    b : array_like
        Denominator
    otherwise : float
        Value to return if `b=0` (default: `np.nan`)
    prec : float
        If `|b|<|prec|` then `otherwise`

    Returns
    -------
    ratio : numpy array or masked array
        *a/b* if *|b| >  |prec|*, *otherwise* if *|b| <= |prec|*

    Notes
    -----
    Output is numpy array. It is a masked array if at least one of *a* or *b*
    is a masked array.

    Examples
    --------
    >>> a = [1., 2., 3.]
    >>> b = 2.
    >>> print('{:.1f} {:.1f} {:.1f}'.format(*division(a, b)))
    0.5 1.0 1.5

    >>> a = [1., 1., 1.]
    >>> b = [2., 1., 0.]
    >>> print(division(a, b))
    [0.5, 1.0, nan]
    >>> print(division(a, b, 0.))
    [0.5, 1.0, 0.0]
    >>> print(division(a, b, otherwise=0.))
    [0.5, 1.0, 0.0]
    >>> print(division(a, b, prec=1.))
    [0.5, nan, nan]

    >>> import numpy as np
    >>> a = np.array([1., 1., 1.])
    >>> b = [2., 1., 0.]
    >>> print(division(a, b))
    [0.5 1.  nan]
    >>> b = np.array([2., 1., 0.])
    >>> print(division(a, b))
    [0.5 1.  nan]

    >>> mask = [0, 0, 1]
    >>> b = np.ma.array([2., 1., 0.], mask=mask)
    >>> print(division(a, b))
    [0.5 1.0 --]

    """
    aa = input2array(a, undef=np.nan, default=0)
    bb = input2array(b, undef=np.nan, default=1)

    bbb = np.where(np.abs(bb) > np.abs(prec), bb, 1)
    out = np.where(np.abs(bb) > np.abs(prec), aa / bbb, otherwise)

    out = array2input(out, a, b, undef=np.nan)

    return out


def div(*args, **kwargs):
    """
    Wrapper function for :func:`division`.

    Examples
    --------
    >>> a = [1., 2., 3.]
    >>> b = 2.
    >>> print(div(a, b))
    [0.5, 1.0, 1.5]

    >>> a = [1., 1., 1.]
    >>> b = [2., 1., 0.]
    >>> print(div(a, b))
    [0.5, 1.0, nan]
    >>> print(div(a, b, 0.))
    [0.5, 1.0,  0.0]
    >>> print(div(a, b, otherwise=0.))
    [0.5, 1.0, 0.0]
    >>> print(div(a, b, prec=1.))
    [0.5, nan, nan]

    """
    return division(*args, **kwargs)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
This is the unittest for esat module.

python -m unittest -v tests/test_air_humidity.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_air_humidity.py

"""
import unittest
import numpy as np
from numpy.ma import masked
import pandas as pd


def _flatten(itr, decimals=0):
    if isinstance(itr, np.ma.MaskedArray):
        fitr = np.ma.around(np.ma.array(itr).flatten(), decimals)
    else:
        fitr = np.around(np.array(itr).flatten(), decimals)
    if len(fitr) == 0:
        return list(fitr)
    else:
        if isinstance(fitr[0], str):
            return [ i for i in fitr ]
        else:
            return list(fitr)


class TestEsat(unittest.TestCase):
    """
    Tests for test_air_humidity.py
    """

    def test_esat(self):
        from pyjams import esat
        import warnings

        T0 = 273.15
        T = np.array([20., -20.])
        res = [2335.847, 103.074]

        # scalar
        T1 = T0 + 20.
        assert isinstance(esat(T1), float)
        assert np.around(esat(T1), 3) == 2335.847
        T1 = T0 - 20.
        assert isinstance(esat(T1), float)
        assert np.around(esat(T1), 3) == 103.074

        # list - different formulas with different word cases
        T1 = [ tt + T0 for tt in T ]
        assert isinstance(esat(T1), list)
        es = esat(T1)
        self.assertEqual(_flatten(es, 3), res)
        formulas = {'GoffGratch': res,
                    'MartiMauersberger': [2335.847, 103.650],
                    'MagnusTeten': [2335.201, 102.771],
                    'buck': [2338.340, 103.286],
                    'Buck_original': [2337.282, 103.267],
                    'wmo': [2337.080, 103.153],
                    'WEXLER': [2323.254, 103.074],
                    'Sonntag': [2339.249, 103.249],
                    'Bolton': [2336.947, 103.074],
                    'Fukuta': [2335.847, 103.074],
                    'HylandWexler': [2338.804, 103.260],
                    'IAPWS': [2339.194, 103.074],
                    'MurphyKoop': [2339.399, 103.252]}
        for ff in formulas:
            es = esat(T1, formula=ff)
            self.assertEqual(_flatten(es, 3), formulas[ff])

        # tuple
        T1 = tuple([ tt + T0 for tt in T ])
        es = esat(T1)
        assert isinstance(es, tuple)
        self.assertEqual(_flatten(es, 3), res)
        # ndarray
        T1 = np.vstack([T, T]) + T0
        res2 = res[:]
        res2.extend(res)
        es = esat(T1)
        assert isinstance(esat(T1), np.ndarray)
        self.assertEqual(_flatten(es, 3), res2)
        # masked_array
        T1 = np.ma.array(T + T0, mask=(T == 20.))
        es = esat(T1)
        assert isinstance(es, np.ndarray)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(_flatten(es, 3), [masked, 103.074])
        # pandas.Series
        T1 = [ tt + T0 for tt in T ]
        d1 = [pd.to_datetime('2020-06-01 12:30'),
              pd.to_datetime('2020-09-03 16:00')]
        df = pd.Series(T1)
        df.index = d1
        es = esat(df)
        assert isinstance(es, pd.Series)
        self.assertEqual(_flatten(es.values, 3), res)
        # pandas.DataFrame
        T1 = np.vstack([T, T]) + T0
        res2 = res[:]
        res2.extend(res)
        d1 = [pd.to_datetime('2020-06-01 12:30'),
              pd.to_datetime('2020-09-03 16:00')]
        df = pd.DataFrame(T1)
        df.index = d1
        es = esat(df)
        assert isinstance(es, pd.DataFrame)
        self.assertEqual(_flatten(es.values.flatten(), 3), res2)

        # liquid
        T1 = [ tt + T0 for tt in T ]
        es = esat(T1, liquid=True)
        self.assertEqual(_flatten(es, 3), [2335.847, 125.292])
        es = esat(T1, formula='Fukuta', liquid=True)
        self.assertEqual(_flatten(es, 3), [2335.847, 125.079])

        # undef, scalar
        T1 = -9998.
        es = esat(T1, undef=-9998.)
        assert isinstance(es, float)
        self.assertEqual(es, -9998.)
        T1 = -9999.
        es = esat(T1)
        assert isinstance(es, float)
        self.assertEqual(es, -9999.)

        # undef, list
        T1 = [20. + T0, -9998.]
        es = esat(T1, undef=-9998.)
        assert isinstance(es, list)
        self.assertEqual(_flatten(es, 3), [2335.847, -9998.])
        T1 = [20. + T0, -9999.]
        es = esat(T1)
        assert isinstance(es, list)
        self.assertEqual(_flatten(es, 3), [2335.847, -9999.])

        # undef, tuple
        T1 = (20. + T0, -9998.)
        es = esat(T1, undef=-9998.)
        assert isinstance(es, tuple)
        self.assertEqual(_flatten(es, 3), [2335.847, -9998.])
        T1 = (20. + T0, -9999.)
        es = esat(T1)
        assert isinstance(es, tuple)
        self.assertEqual(_flatten(es, 3), [2335.847, -9999.])

        # undef, ndarray
        T1 = np.array([20. + T0, -9998.])
        es = esat(T1, undef=-9998.)
        assert isinstance(es, np.ndarray)
        self.assertEqual(_flatten(es, 3), [2335.847, -9998.])
        T1 = np.array([20. + T0, -9999.])
        es = esat(T1)
        assert isinstance(es, np.ndarray)
        self.assertEqual(_flatten(es, 3), [2335.847, -9999.])

        # undef, masked array
        T = np.array([20. + T0, -9998., -9999.])
        T1 = np.ma.array(T, mask=(T == -9999.))
        es = esat(T1, undef=-9998.)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(_flatten(es, 3), [2335.847, masked, masked])
        T = np.array([20. + T0, -9998., -9999.])
        T1 = np.ma.array(T, mask=(T == -9998.))
        es = esat(T1)
        assert isinstance(es, np.ma.MaskedArray)
        self.assertEqual(_flatten(es, 3), [2335.847, masked, masked])

        # undef and nan, pandas.Series
        T = [20., -20., -9999. - T0, np.nan]
        T1 = [ tt + T0 for tt in T ]
        d1 = [pd.to_datetime('2020-06-01 12:30'),
              pd.to_datetime('2020-09-03 16:00'),
              pd.to_datetime('2021-06-01 12:30'),
              pd.to_datetime('2021-09-03 16:00')]
        df = pd.Series(T1)
        df.index = d1
        es = esat(df, undef=-9999.)
        assert isinstance(es, pd.Series)
        assert np.isnan(es.iloc[-1])
        self.assertEqual(_flatten(es.values[:-1], 3),
                         [2335.847, 103.074, -9999.])

        # undef and nan, pandas.DataFrame
        T = [20., -20., -9999. - T0, np.nan]
        T1 = np.array([T, T]) + T0
        d1 = [pd.to_datetime('2020-06-01 12:30'),
              pd.to_datetime('2020-09-03 16:00')]
        df = pd.DataFrame(T1)
        df.index = d1
        es = esat(df, undef=-9999.)
        assert isinstance(es, pd.DataFrame)
        assert np.isnan(es.iloc[-1, -1])
        self.assertEqual(_flatten(es.values[:, :-1], 3),
                         [2335.847, 103.074, -9999.,
                          2335.847, 103.074, -9999.])

        # errors
        # T < 0
        T1 = T0 - T0
        self.assertRaises(AssertionError, esat, T1)
        # T > 373.15
        T1 = T0 + 101.
        self.assertRaises(AssertionError, esat, T1)
        # wrong formula
        T1 = T0 + 20.
        self.assertRaises(ValueError, esat, T1, formula='dummy')
        T1 = T0 - 20.
        self.assertRaises(ValueError, esat, T1, formula='dummy')

        # warning T < 100
        T1 = T0 - 200.
        with warnings.catch_warnings(record=True) as w:
            es = esat(T1)
            assert len(w) > 0

    def test_air2air(self):
        from pyjams import eair2rhair
        from pyjams import rhair2eair
        from pyjams import eair2vpd
        from pyjams import vpd2eair
        from pyjams import rhair2vpd
        from pyjams import vpd2rhair
        from pyjams import eair2shair
        from pyjams import shair2eair
        from pyjams import eair2mrair
        from pyjams import mrair2eair

        tests = ['eair2rhair', 'rhair2eair',
                 'eair2vpd', 'vpd2eair',
                 'rhair2vpd', 'vpd2rhair',
                 'eair2shair', 'shair2eair',
                 'eair2mrair', 'eair2mrair_mol',
                 'mrair2eair', 'mrair2eair_mol']
        for test in tests:
            if test == 'eair2rhair':
                func = eair2rhair
                xair = 1000.
                Ta = 293.15
                res2 = 42.81
                fac = 100.
                kwargs = {}
            elif test == 'rhair2eair':
                func = rhair2eair
                xair = 0.5
                Ta = 293.15
                res2 = 1167.92
                fac = 1.
                kwargs = {}
            elif test == 'eair2vpd':
                func = eair2vpd
                xair = 1000.
                Ta = 293.15
                res2 = 1335.85
                fac = 1.
                kwargs = {}
            elif test == 'vpd2eair':
                func = vpd2eair
                xair = 1000.
                Ta = 293.15
                res2 = 1335.85
                fac = 1.
                kwargs = {}
            elif test == 'rhair2vpd':
                func = rhair2vpd
                xair = 0.5
                Ta = 293.15
                res2 = 1167.92
                fac = 1.
                kwargs = {}
            elif test == 'vpd2rhair':
                func = vpd2rhair
                xair = 1000.
                Ta = 293.15
                res2 = 57.19
                fac = 100.
                kwargs = {}
            elif test == 'eair2shair':
                func = eair2shair
                xair = 1000.
                Ta = 101325.
                res2 = 6.16
                fac = 1000.
                kwargs = {}
            elif test == 'shair2eair':
                func = shair2eair
                xair = 0.006
                Ta = 101325.
                res2 = 973.89
                fac = 1.
                kwargs = {}
            elif test == 'eair2mrair':
                func = eair2mrair
                xair = 1000.
                Ta = 101325.
                res2 = 6.2
                fac = 1000.
                kwargs = {}
            elif test == 'eair2mrair_mol':
                func = eair2mrair
                xair = 1000.
                Ta = 101325.
                res2 = 9.97
                fac = 1000.
                kwargs = {'mol': True}
            elif test == 'mrair2eair':
                func = mrair2eair
                xair = 0.006
                Ta = 101325.
                res2 = 968.10
                fac = 1.
                kwargs = {}
            elif test == 'mrair2eair_mol':
                func = mrair2eair
                xair = 0.006
                Ta = 101325.
                res2 = 604.32
                fac = 1.
                kwargs = {'mol': True}

            # scalar
            e1 = xair
            T1 = Ta
            yair = func(e1, T1, **kwargs) * fac
            assert isinstance(yair, float)
            assert np.around(yair, 2) == res2

            # list
            e1 = [xair, xair]
            T1 = Ta
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, list)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            e1 = xair
            T1 = [Ta, Ta]
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, list)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            e1 = [xair, xair]
            T1 = [Ta, Ta]
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, list)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            # tuple
            e1 = tuple([xair, xair])
            T1 = tuple([Ta, Ta])
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, tuple)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            # ndarray
            e1 = np.array([xair, xair])
            T1 = np.array([Ta, Ta])
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, np.ndarray)
            yair *= fac
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            e1 = np.array([[xair, xair], [xair, xair]])
            T1 = np.array([[Ta, Ta], [Ta, Ta]])
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, np.ndarray)
            yair *= fac
            self.assertEqual(_flatten(yair, 2), [res2, res2, res2, res2])

            # masked_array
            e1 = [xair, xair]
            T1 = [Ta, Ta]
            e1 = np.ma.array(e1)
            T1 = np.ma.array(T1)
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, np.ndarray)
            assert isinstance(yair, np.ma.MaskedArray)
            yair *= fac
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            # mixed types
            e1 = [xair, xair]
            T1 = tuple([Ta, Ta])
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, list)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            e1 = tuple([xair, xair])
            T1 = [Ta, Ta]
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, tuple)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            e1 = np.array(e1)
            T1 = [Ta, Ta]
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, np.ndarray)
            yair *= fac
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            e1 = [xair, xair]
            T1 = np.array(T1)
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, list)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, res2])

            # pandas.Series
            e1 = np.array([xair, xair])
            T1 = np.array([Ta, Ta])
            d1 = [pd.to_datetime('2020-06-01 12:30'),
                  pd.to_datetime('2020-09-03 16:00')]
            d2 = [pd.to_datetime('2021-06-01 12:30'),
                  pd.to_datetime('2021-09-03 16:00')]
            df1 = pd.Series(e1)
            df1.index = d1
            df2 = pd.Series(T1)
            df2.index = d2
            yair = func(df1, df2, **kwargs)
            assert isinstance(yair, pd.Series)
            yair *= fac
            self.assertEqual(_flatten(yair.values, 2), [res2, res2])

            # pandas.DataFrame
            e1 = np.array([xair, xair])
            T1 = np.array([Ta, Ta])
            d1 = [pd.to_datetime('2020-06-01 12:30'),
                  pd.to_datetime('2020-09-03 16:00')]
            df = pd.DataFrame({'xair': e1, 'Ta': T1})
            df.index = d1
            yair = func(df['xair'], df['Ta'], **kwargs)
            assert isinstance(yair, pd.Series)
            yair *= fac
            self.assertEqual(_flatten(yair.values, 2), [res2, res2])

            # undef, scalar
            e1 = -9998.
            T1 = Ta
            es = func(e1, T1, undef=-9998., **kwargs)
            assert isinstance(es, float)
            assert np.around(es, 2) == -9998.
            e1 = -9999.
            T1 = Ta
            es = func(e1, T1, **kwargs)
            assert isinstance(es, float)
            assert np.around(es, 2) == -9999.
            e1 = xair
            T1 = -9999.
            es = func(e1, T1, **kwargs)
            assert isinstance(es, float)
            assert np.around(es, 2) == -9999.

            # undef, list
            e1 = [xair, -9999.]
            T1 = [Ta, Ta]
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, list)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, -9999. * fac])

            e1 = [xair, xair]
            T1 = [Ta, -9999.]
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, list)
            yair = [ ee * fac for ee in yair ]
            self.assertEqual(_flatten(yair, 2), [res2, -9999. * fac])

            # undef, masked array
            e1 = np.ma.array([xair, -9998., -9999.])
            T1 = [Ta, Ta, Ta]
            e1 = np.ma.array(e1, mask=(e1 == -9998.))
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, np.ma.MaskedArray)
            yair *= fac
            self.assertEqual(_flatten(yair, 2), [res2, masked, masked])

            e1 = [xair, xair, xair]
            T1 = np.ma.array([Ta, -9998., -9999.])
            T1 = np.ma.array(T1, mask=(T1 == -9998.))
            yair = func(e1, T1, **kwargs)
            assert isinstance(yair, np.ma.MaskedArray)
            yair *= fac
            self.assertEqual(_flatten(yair, 2), [res2, masked, masked])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
Kinetic fractionation factors for molecular diffusion of water isotopologues.

This module was written by Matthias Cuntz while at Department of Computational
Hydrosystems, Helmholtz Centre for Environmental Research - UFZ, Leipzig,
Germany, and continued while at Institut National de Recherche pour
l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2014-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   alpha_kin_h2o

History
    * Written, Sep 2014, Matthias Cuntz
    * Code refactoring, Nov 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""


__all__ = ['alpha_kin_h2o']


def alpha_kin_h2o(isotope=None, eps=False, greater1=True,
                  boundary=False, cappa=False):
    """
    Kinetic fractionation factors for molecular diffusion of water
    isotopologues.

    It does not use the atmospheric convention, i.e. factor < 1, by default but
    sets factor > 1 (greater1=True).

    Parameters
    ----------
    isotope : int, optional
        Select water isotopologue: 1: HDO; 2: H218O; else: no fractionation,
        i.e. return 1 (default)
    eps : bool, optional
        Returns fractionation epsilon=alpha-1 instead of fractionation factor
        alpha if True (default: return alpha)
    greater1 : bool, optional
        alpha > 1 if True, which is not the atmospheric convention.
        alpha < 1 if False, which is the atmospheric convention.
    boundary : bool, optional
        Returns `alpha**2/3` for diffusion through boundary layer instead of
        molecular diffusion if True (default: False).
    cappa : bool, optional
        Uses factors of Cappa et al. (2003) instead of Merlivat (1978) if True
        (default: False).

    Returns
    -------
    alpha / epsilon : float or array-like
        Kinetic fractionation factor (alpha) or fractionation (epsilon)

    Notes
    -----
    Cappa, C. D., Hendricks, M. B., DePaolo, D. J., & Cohen, R. (2003)
        Isotopic fractionation of water during evaporation
        Journal of Geophysical Research, 108(D16), 4525.
        doi:10.1029/2003JD003597
    Merlivat, L. (1978)
        Molecular Diffusivities Of (H2O)-O-16 HD16O, And (H2O)-O-18 In Gases,
        The Journal of Chemical Physics, 69(6), 2864-2871.
    Merlivat, L., & Jouzel, J. (1979)
        Global climatic interpretation of the deuterium-oxygen-18 relationship
        for precipitation, Journal of Geophysical Research, 84(C8), 5029-5033.

    Examples
    --------
    Fractionation factor

    >>> import numpy as np
    >>> print(np.around(alpha_kin_h2o(isotope=0), 4))
    1.0

    Fractionations

    >>> print(np.around(alpha_kin_h2o(isotope=1, eps=True)*1000., 4))
    25.1153
    >>> print(np.around(alpha_kin_h2o(isotope=2, eps=True,
    ...                               greater1=False)*1000., 4))
    -27.3
    >>> print(np.around(alpha_kin_h2o(isotope=2, eps=True, greater1=True,
    ...                               boundary=True)*1000., 4))
    18.6244
    >>> print(np.around(alpha_kin_h2o(isotope=2, eps=True, greater1=False,
    ...                               boundary=True, cappa=True)*1000., 4))
    -20.7076

    """
    # Fractionation factors
    if (isotope == 1):  # HDO
        if cappa:
            out = 0.9839
        else:
            out = 0.9755
    elif (isotope == 2):  # H218O
        if cappa:
            out = 0.9691
        else:
            out = 0.9727
    else:
        out = 1.

    # boundary layer
    if boundary:
        out = out**(2./3.)

    # alpha+
    if greater1:
        out = 1./out

    # epsilon
    if eps:
        out -= 1.

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
import os

def readhdf(fName, var='', reform=False, squeeze=False, variables=False,
            attributes=False, fileattributes=False, sort=False):
    """
        Get variables or print information of hdf4 or hdf5 file.


        Definition
        ----------
        def readhdf(fName, var='', reform=False, squeeze=False, variables=False,
                    attributes=False, fileattributes=False, sort=False):

        Input
        -----
        fName            hdf4/5 file name


        Optional Input Parameters
        -------------------------
        var              name of variable in hdf file


        Options
        -------
        reform           if output is array then squeeze(array)
        squeeze          same as reform
        variables        get list of variables in hdf file
        attributes       get dictionary of all attributes of specific variable
        fileattributes   get dictionary of all attributes of the file
        sort             sort variable names.


        Output
        ------
        Either variable array or information of file aor variable
        such as list of all variables or dictionary of variable attributes.


        Examples
        --------
        # Do only HDF5 tests. HDF4 tests are in wrapper function hdfread
        # so that al least the HDF5 doctests can be run for Python 3.
        # HDF5
        >>> a = readhdf('test_readhdf5.hdf5', fileattributes=True)
        >>> print(a['NB_PARAMETERS'])
        9

        >>> print([ str(i) for i in readhdf('test_readhdf5.hdf5', variables=True) ])
        ['chs']

        >>> a = readhdf('test_readhdf5.hdf5', var='chs', attributes=True)
        >>> print([ str(i) for i in sorted(list(a.keys())) ])
        ['Double', 'Inttest', 'LLLLLL', 'What the hell']
        >>> print(a['Double'])
        [1.1]

        >>> from autostring import astr
        >>> print(astr(readhdf('test_readhdf5.hdf5', var='chs'),3,pp=True))
        [['  1.000' '  2.000' '  3.000' '  3.000' '  2.000']
         ['  1.000' ' 23.000' '254.000' '  5.000' '  4.654']
         ['654.654' ' 54.540' '546.540' '564.546' '  5.500']
         ['  1.100' '  2.200' '  0.000' '  3.300' '  4.400']
         ['  0.000' '  1.000' '  2.000' '  3.000' '  0.000']]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2013 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jun 2012
        Modified, MC, Feb 2013 - ported to Python 3
    """
    # Open hdf5 file
    if not os.path.isfile(fName):
        raise IOError('File does not exist: '+fName)

    try:
        from pyhdf.SD import SD
        haveh4 = True
        from readhdf4 import readhdf4
    except ImportError:
        haveh4 = False

    try:
        import h5py as hdf5
        haveh5 = True
        from readhdf5 import readhdf5
    except ImportError:
        haveh5 = False

    if haveh5:
        try:
            f = hdf5.File(fName, 'r')
            f.close()
            return readhdf5(fName, var=var, reform=reform, squeeze=squeeze, variables=variables,
                            attributes=attributes, fileattributes=fileattributes, sort=sort)
        except:
            if haveh4:
                try:
                    f = SD(fName)
                    f.end()
                    return readhdf4(fName, var=var, reform=reform, squeeze=squeeze, variables=variables,
                                attributes=attributes, fileattributes=fileattributes, sort=sort)
                except:
                    raise IOError('File is not in hdf4 nor in hdf5 format: '+fName)
            else:
                raise IOError('File is not in hdf5 format: '+fName)
    else:
        if haveh4:
            try:
                f = SD(fName)
                f.end()
                return readhdf4(fName, var=var, reform=reform, squeeze=squeeze, variables=variables,
                                attributes=attributes, fileattributes=fileattributes, sort=sort)
            except:
                raise IOError('File is not in hdf4 format: '+fName)



def hdfread(*args, **kwargs):
    """
        Wrapper for readhdf.
        def readhdf(fName, var='', reform=False, squeeze=False, variables=False,
                    attributes=False, fileattributes=False, sort=False):


        Examples
        --------
        # HDF4
        >>> var = hdfread('test_readhdf4.hdf4', fileattributes=True)
        >>> print(list(var.keys()))
        ['OldCoreMetadata.0', 'HDFEOSVersion', 'OldArchiveMetadata.0', 'OldStructMetadata.0', 'StructMetadata.0']
        >>> print(var['HDFEOSVersion'])
        ('HDFEOS_V2.14', 0, 4, 12)

        >>> var = hdfread('test_readhdf4.hdf4', variables=True)
        >>> print(var)
        ['QC_250m_1', 'sur_refl_b02_1', 'sur_refl_b01_1', 'num_observations']

        >>> var = hdfread('test_readhdf4.hdf4', variables=True, sort=True)
        >>> print(var)
        ['QC_250m_1', 'num_observations', 'sur_refl_b01_1', 'sur_refl_b02_1']

        >>> var = hdfread('test_readhdf4.hdf4', var='sur_refl_b01_1')
        >>> print(var)
        [[7492 7327 7327 7131 7187]
         [6604 6604 7423 7131 7131]
         [7441 7441 7423 7423 7507]]

        >>> var = hdfread('test_readhdf4.hdf4', var='sur_refl_b01_1', attributes=True)
        >>> print(list(var.keys()))
        ['_FillValue', 'Nadir Data Resolution', 'scale_factor', 'valid_range', 'add_offset', 'long_name', 'calibrated_nt', 'units', 'scale_factor_err', 'add_offset_err', 'HorizontalDatumName']
        >>> print(var['_FillValue'])
        (-28672, 3, 22, 1)
    """
    return readhdf(*args, **kwargs)


if __name__ == '__main__':
    import doctest
    try:
        from pyhdf.SD import SD
        doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
    except ImportError:
        doctest.run_docstring_examples(readhdf, None, optionflags=doctest.NORMALIZE_WHITESPACE)

    # print readhdf('test_readhdf4.hdf4', variables=True)
    # print readhdf('test_readhdf5.hdf5', variables=True)




#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def convex_hull(points, graphic=False, smidgen=0.0075):
    """
        Calculate subset of 2D points that make a convex hull around a set of
        2D points. Recursively eliminates points that lie inside two
        neighbouring points until only convex hull is remaining.


        Definition
        ----------
        def convex_hull(points, graphic=True, smidgen=0.0075)


        Input
        -----
        points       ndarray (2 x m), array of points for which to find hull


        Optional Input
        --------------
        graphic      bool, use pylab to show progress
        smidgen      float, offset for graphic number labels - useful
                     values depend on your data range


        Output
        ------
        hull_points  ndarray (2 x n), convex hull surrounding points


        References
        ----------
        This routine (with additional subroutines) was coded originally by
        Angus McMorland, 2007.
        It is a copy from the scipy cookbook page:
            http://www.scipy.org/Cookbook/Finding_Convex_Hull


        Examples
        --------
        # make some points
        >>> points = np.array([[2,3,2,4,5,5,7,5,5],[1,2,4,3,6,4,3,2,1]])
        >>> hull_xy = convex_hull(points, graphic=False, smidgen=0.075)
        >>> from autostring import astr
        >>> print(astr(hull_xy,pp=True))
        [['5' '1']
         ['7' '3']
         ['5' '6']
         ['2' '4']
         ['2' '1']]

        >>> hull_xy = convex_hull(points, graphic=False, smidgen=0.075)
        >>> print(astr(hull_xy,pp=True))
        [['5' '1']
         ['7' '3']
         ['5' '6']
         ['2' '4']
         ['2' '1']]

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2009-2013 Arndt Piayda

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Nov 2012
        Modified, AP, Dec 2012 - documentation change
        Modified, MC, Feb 2013 - ported to Python 3
    """

    if graphic:
        import pylab as p
        p.clf()
        p.plot(points[0], points[1], 'ro')
    n_pts = points.shape[1]
    assert(n_pts > 5)
    centre = points.mean(1)
    if graphic: p.plot((centre[0],),(centre[1],),'bo')
    angles = np.apply_along_axis(_angle_to_point, 0, points, centre)
    pts_ord = points[:,angles.argsort()]
    if graphic:
        for i in range(n_pts):
            p.text(pts_ord[0,i] + smidgen, pts_ord[1,i] + smidgen, '%d' % i)
    pts = [x[0] for x in zip(pts_ord.transpose())]
    prev_pts = len(pts) + 1
    k = 0
    while prev_pts > n_pts:
        prev_pts = n_pts
        n_pts = len(pts)
        if graphic: p.gca().patches = []
        i = -2
        while i < (n_pts - 2):
            Aij = area_of_triangle(centre, pts[i],     pts[(i + 1) % n_pts])
            Ajk = area_of_triangle(centre, pts[(i + 1) % n_pts], pts[(i + 2) % n_pts])
            Aik = area_of_triangle(centre, pts[i],     pts[(i + 2) % n_pts])
            if graphic:
                _draw_triangle(centre, pts[i], pts[(i + 1) % n_pts], graphic, facecolor='blue', alpha = 0.2)
                _draw_triangle(centre, pts[(i + 1) % n_pts], pts[(i + 2) % n_pts], graphic, facecolor='green', alpha = 0.2)
                _draw_triangle(centre, pts[i], pts[(i + 2) % n_pts], graphic, facecolor='red', alpha = 0.2)
            if Aij + Ajk < Aik:
                if graphic: p.plot((pts[i + 1][0],),(pts[i + 1][1],),'go')
                del pts[i+1]
            i += 1
            n_pts = len(pts)
        k += 1
    if graphic: p.show()
    return np.asarray(pts)

def _angle_to_point(point, centre):
    '''calculate angle in 2-D between points and x axis'''
    delta = point - centre
    res = np.arctan(delta[1] / delta[0])
    if delta[0] < 0:
        res += np.pi
    return res

def _draw_triangle(p1, p2, p3, graphic, **kwargs):
    if graphic:
        import pylab as p
    tmp = np.vstack((p1,p2,p3))
    x,y = [x[0] for x in zip(tmp.transpose())]
    p.fill(x,y, **kwargs)

def area_of_triangle(p1, p2, p3):
    '''calculate area of any triangle given co-ordinates of the corners'''
    return np.linalg.norm(np.cross((p2 - p1), (p3 - p1)))/2.

if __name__ == "__main__":
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)




#!/usr/bin/env python
"""
Prints arguments on screen and in file, like Unix/Linux tee utility.

This module was written by Matthias Cuntz while at Department of Computational
Hydrosystems, Helmholtz Centre for Environmental Research - UFZ, Leipzig,
Germany, and continued while at Institut National de Recherche pour
l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2014-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided:

.. autosummary::
   tee

History
    * Written Oct 2014 by Matthias Cuntz (mc (at) macu (dot) de)
    * Allow that file keyword not given, Nov 2016, Matthias Cuntz
    * Make numpy doctsring format, Dec 2019, Matthias Cuntz
    * Code refactoring, Sep 2021, Matthias Cuntz
    * More consistent docstrings, Jan 2022, Matthias Cuntz

"""


__all__ = ['tee']


def tee(*args, **kwargs):
    """
    Prints arguments on screen and in file, like Unix/Linux tee utility.

    Calls print function twice, once with the keyword file and once without,
    i.e. prints on sys.stdout.

    Parameters
    ----------
    *args : iterable
        All arguments of the print function; will be passed to the `print`
        function.
    **kwargs : dict
        All keyword arguments of the `print` function; will be passed to the
        `print` function.
    file : object
        The `file` argument must be an object with a `write(string)` method. If
        it is not present or None, `*args` will be printed on sys.stdout only.
        Since printed arguments are converted to text strings, `print()` cannot
        be used with binary mode file objects.arguments.

    Returns
    -------
    None
        If `file` is given and not None, then print will be called with `*args`
        and `**kwargs` and a second time with the `file` keyword argument
        removed, so that `*args` will be written to sys.stdout.
        If `file` is not given or None, `*args` will only be written to
        sys.stdout, i.e. *tee* is a simple wrapper of the `print` function.

    Examples
    --------
    >>> st = 'Output on screen and in log file'
    >>> ff = 'tee_log.txt'
    >>> f = open(ff, 'w')
    >>> tee(st, file=f)
    Output on screen and in log file
    >>> f.close()

    >>> f = open(ff, 'r')
    >>> test = f.readline()
    >>> f.close()
    >>> print(test[:-1])  # rm trailing newline character
    Output on screen and in log file
    >>> import os
    >>> os.remove(ff)

    >>> f=None
    >>> st = 'Output only on screen'
    >>> tee(st, file=f)
    Output only on screen
    >>> tee(st)
    Output only on screen

    """
    if 'file' in kwargs:
        if kwargs['file'] is not None:
            print(*args, **kwargs)  # file
        del kwargs['file']
    print(*args, **kwargs)          # screen


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
pyjams is a general Python package with a wide variety of miscellaneous utility
functions. It is offering miscellaneous functions in different categories, such
as reading different file formats, date conversion routines, or meteorological
functions. It has several subpackages offering constants, special functions as
well as objective functions to be used with scipy.optimize.fmin or
scipy.optimize.curvefit, and much more.

The package has evolved from its predecessor the JAMS Python package
https://github.com/mcuntz/jams_python

:copyright: Copyright 2021-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

Subpackages
===========
.. autosummary::
   air_humidity
   alpha_equ_h2o
   alpha_kin_h2o
   argsort
   gridcellarea
   closest
   color
   const
   date2date
   class_datetime
   division
   fgui
   fsread
   functions
   jams
   kernel_regression
   mad
   mcplot
   means
   morris_method
   ncio
   npyio
   pack
   position
   readnetcdf
   romanliterals
   sce
   screening
   str2tex
   tee
   text2plot

History
   * Written Oct 2021 by Matthias Cuntz (mc (at) macu (dot) de)
   * v1.0, initial Github, PyPI, Zenodo commit, Oct 2021, Matthias Cuntz
   * v1.1, automatic versioning, zenodo defaults, Oct 2021, Matthias Cuntz
   * v1.2, added closest, Oct 2021, Matthias Cuntz
   * v1.3, added argsort, Oct 2021, Matthias Cuntz
   * v1.4, added division, Oct 2021, Matthias Cuntz
   * v1.5, added alpha_equ_h2o, Oct 2021, Matthias Cuntz
   * v1.6, added alpha_kin_h2o, Nov 2021, Matthias Cuntz
   * v1.7, added mcPlot, Nov 2021, Matthias Cuntz
   * v1.8, added str2tex and color, Nov 2021, Matthias Cuntz
   * v1.9, added position, Nov 2021, Matthias Cuntz
   * v1.10, added sron colors, Nov 2021, Matthias Cuntz
   * v1.11, added text2plot, abc2plot, signature2plot,
     Nov 2021, Matthias Cuntz
   * v1.12, added date2dec and all its wrappers, Dec 2021, Matthias Cuntz
   * v1.13, added fsread, fread, sread, Dec 2021, Matthias Cuntz
   * v1.14, added GUI dialogs to choose files and directories using Tkinter,
     Jan 2022, Matthias Cuntz
   * v1.15, added esat, Jan 2022, Matthias Cuntz
   * v1.16, added mad, Jan 2022, Matthias Cuntz
   * v1.17, added xread, xlsread, and xlsxread, Jan 2022, Matthias Cuntz
   * v1.18, helper module to preserve input types, Mar 2022, Matthias Cuntz
   * v1.19, added readnetcdf, Mar 2022, Matthias Cuntz
   * v1.20, added gridcellarea and kernel_regression, Apr 2022, Matthias Cuntz
   * v1.21, enhancements in color and kernel_regression,
     Apr 2022, Matthias Cuntz
   * v1.22, added ncio, May 2022, Matthias Cuntz
   * v1.23, added datetime module, Jun 2022, Matthias Cuntz
   * v1.24, moved docu to Github Pages, Jun 2022, Matthias Cuntz
   * v1.25, Microseconds and negative years in date2date and datetime,
     Jun 2022, Matthias Cuntz
   * v1.26, added means, Jul 2022, Matthias Cuntz
   * v1.27, added sce, Dec 2022, Matthias Cuntz
   * v1.28, added updatez, Jan 2023, Matthias Cuntz
   * v1.29, added functions for converting humidity in air,
     Jan 2023, Matthias Cuntz
   * v1.30, hvplot in mcplot, May 2023, Matthias Cuntz
   * v1.31, added pack and unpack, Jun 2023, Matthias Cuntz
   * v1.32, added jams, Jun 2023, Matthias Cuntz
   * v1.33, start supporting pandas Series and DataFrames,
     Jul 2023, Matthias Cuntz

"""
# version, author
try:  # pragma: no cover
    from ._version import __version__
except ImportError:  # pragma: no cover
    # package is not installed
    __version__ = "0.0.0.dev0"
__author__  = "Matthias Cuntz, Juliane Mai, Stephan Thober, Arndt Piayda"

# sub-packages without dependencies to rest of pyjams
# color palettes and continuous color maps
from . import color
# physical, mathematical, computational, isotope, and material constants
from . import const
# variety of specialised functions
from . import functions
# netCDF4 functions to copy netcdf file while doing some transformations on
# variables and dimensions.
from . import ncio
# old JAMS routines
from . import jams

# air humidity calculations
from .air_humidity import esat, eair2rhair, rhair2eair
from .air_humidity import eair2vpd, vpd2eair, rhair2vpd, vpd2rhair
from .air_humidity import eair2shair, shair2eair, eair2mrair, mrair2eair
# isotopic fractionation factors during liquid-water vapour equilibration
from .alpha_equ_h2o import alpha_equ_h2o
# kinetic fractionation of molecular diffusion of water vapour
from .alpha_kin_h2o import alpha_kin_h2o
# argmax, argmin and argsort for array_like and Python iterables
from .argsort import argmax, argmin, argsort
# Area of grid cells on Earth
from .gridcellarea import gridcellarea
# (index of) closest element in an array
from .closest import closest
# convert date representations between different regional variants
from .date2date import date2date
from .date2date import date2en, date2fr, date2us
from .date2date import en2date, en2fr, en2us
from .date2date import fr2date, fr2en, fr2us
from .date2date import us2date, us2en, us2fr
# catch division by zero
from .division import division, div
# cftime extension
from .class_datetime import date2dec, date2num, dec2date, num2date, datetime
# GUI dialogs to choose files and directories using Tkinter
try:  # Tkinter must be installed
    from .fgui import directory_from_gui, directories_from_gui
    from .fgui import file_from_gui, files_from_gui
except ImportError:
    pass
# read numbers and strings from a file into 2D float and string arrays
from .fsread import fsread, fread, sread
from .fsread import xread, xlsread, xlsxread
# multi-dimensional non-parametric kernel regression
from .kernel_regression import kernel_regression_h, kernel_regression
# median absolute deviation test
from .mad import mad
# Matthias Cuntz' standard plotting class.
from .mcplot import mcPlot
# daily, monthly, yearly, etc. means
from .means import means
# has to be ordered for import: morris -> screening
# Sampling of optimised trajectories for and calculation of Morris Measures /
# Elementary Effects
from .morris_method import morris_sampling, elementary_effects
# update arrays in a single file in numpy's npz format
from .npyio import updatez, updatez_compressed
# like Fortran pack/unpack
from .pack import pack, unpack
# positions of subplots, used with add_axes
from .position import position
# get variables from or print information of a netcdf file
from .readnetcdf import infonetcdf, ncinfo, readnetcdf, ncread
# Convert integer to and from Roman numerals
from .romanliterals import int2roman, roman2int
# Shuffled-Complex-Evolution (SCE) algorithm for function minimization
from .sce import sce
# Sample trajectories, run model and return Morris Elementary Effects
from .screening import screening, ee
# Convert strings to LaTeX strings
from .str2tex import str2tex
# like unix tee
from .tee import tee
# put text on plot (import after str2tex)
from .text2plot import text2plot, abc2plot, signature2plot


__all__ = ['__version__', '__author__',
           'color', 'const', 'functions',
           'esat',
           'eair2rhair', 'rhair2eair',
           'eair2vpd', 'vpd2eair',
           'rhair2vpd', 'vpd2rhair',
           'eair2shair', 'shair2eair',
           'eair2mrair', 'mrair2eair',
           'alpha_equ_h2o', 'alpha_kin_h2o',
           'argmax', 'argmin', 'argsort',
           'gridcellarea',
           'closest',
           'date2date',
           'date2en', 'date2fr', 'date2us',
           'en2date', 'en2fr', 'en2us',
           'fr2date', 'fr2en', 'fr2us',
           'us2date', 'us2en', 'us2fr',
           'division', 'div',
           'date2dec', 'date2num', 'dec2date', 'num2date', 'datetime',
           'directory_from_gui', 'directories_from_gui',
           'file_from_gui', 'files_from_gui',
           'fsread', 'fread', 'sread',
           'xread', 'xlsread', 'xlsxread',
           'kernel_regression_h', 'kernel_regression'
           'mad',
           'mcPlot',
           'means',
           'morris_sampling', 'elementary_effects',
           'updatez', 'updatez_compressed',
           'pack', 'unpack',
           'position',
           'infonetcdf', 'ncinfo', 'readnetcdf', 'ncread',
           'int2roman', 'roman2int',
           'sce',
           'screening', 'ee',
           'str2tex',
           'tee',
           'text2plot', 'abc2plot', 'signature2plot',
           ]



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def area_poly(x, y):
    """
        Determines the are of a polygon represented as a list of (x,y) vertex coordinates,
        implicitly wrapping around from the last vertex to the first.


        Definition
        ----------
        def area_poly(x, y):


        Input
        -----
        x   np.array, x coordinates of the polygon
        y   np.array, y coordinates of the polygon


        Output
        ------
        area


        References
        ----------
        Code from http://stackoverflow.com/questions/451426/how-do-i-calculate-the-surface-area-of-a-2d-polygon
        From the website:
          It is an application of Green's theorem (http://en.wikipedia.org/wiki/Green%27s_theorem#Area_Calculation)
          for the functions -y and x; exactly in the way a planimeter works.
          More specifically:
            Formula above = integral_perimeter(-y dx + x dy) = integral_area((-(-dy)/dy+dx/dx)dydyx = 2*Area


        Example
        --------
        >>> x = np.array([1.0, 2.0, 2.0, 1.0])
        >>> y = np.array([1.0, 1.0, 2.0, 2.0])
        >>> from autostring import astr
        >>> print(astr(area_poly(x,y),1,pp=True))
        1.0


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2019 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Nov 2012 - stackoverflow.com
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, May 2019 - np.sum needs iterable instead of generator in Python 3
    """

    # Could include some checks here
    p = list(zip(x,y))
    return 0.5 * np.abs(np.sum([ x0*y1 - x1*y0 for ((x0, y0), (x1, y1)) in segments(p) ]))


def segments(p):
    return list(zip(p, p[1:] + [p[0]]))


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # x = np.array([1.0, 2.0, 2.0, 1.0])
    # y = np.array([1.0, 1.0, 2.0, 2.0])
    # print(area_poly(x,y))




#!/usr/bin/env python
import numpy as np
import warnings as wa
from pyjams.const import eps


def fftngo(t, y, nbins=False, plot=False):
    """
        Fast fourier transformation for dummies (like me). Takes time and
        amplitude arrays and calculates frequency spectrum.


        Definition
        ----------
        def fftngo(t, y, nbins=False, plot=False):


        Input
        -----
        t            numpy.ndarray(N), with independent values (e.g. time or space)
        y            numpy.ndarray(N), with dependent values (e.g. amplitude)


        Optional Input
        --------------
        nbins        int if not False, number of bins to equally bin the frequency
                     spectrum on a logatithmic scale (default: False, no binning).
                     if not False, output caontains binned frequency spectrum additionaly
                     to original spectrum.
        plot         bool, creates a plot of input and output if True (default: False)


        Output
        ------
        X            numpy.ndarray(N/2), frequency spectrum
        Y            numpy.ndarray(N/2), amplitude spectrum
        binm         if nbins not False, numpy.ndarray(nbins-1), bin means
        mv           if nbins not False, numpy.ndarray(nbins-1), bin values


        Examples
        --------
        >>> from jams import around
        >>> # generate example data
        >>> Ts = 0.5 # [h]               # sampling interval of half hour
        >>> Fs = 1./Ts # [1/h]           # sampling frequency
        >>> t = np.arange(0, 365*24, Ts) # one year time data

        >>> tt = 24. # [h]           # period time of observed phenomena
        >>> ff = 1./tt # [1/h]       # frequency of observed phenomena
        >>> s = np.sin(2*np.pi*t*ff) # one year of phenomena data

        >>> # calculate frequency spectrum
        >>> X, Y = fftngo(t, s)
        
        >>> # get dominant frequency
        >>> fd = X[np.argmax(Y)]
        
        >>> # period equals 24 h
        >>> print(int(1./fd))
        24
        
        >>> # get binned frequency spectrum additionally to standard output
        >>> X, Y, binm, mv = fftngo(t, s, nbins=50)

        
        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Arndt Piayda

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Oct 2016
                  MC, Nov 2016 - const.tiny -> const.eps
    """
    # input check
    assert isinstance(t, np.ndarray) & isinstance(y, np.ndarray), 'x and y must be numpy arrays'
    assert (t.ndim == 1) & (y.ndim == 1),                         'x and y must be 1d arrays'
    assert (~np.isnan(t).any()) & (~np.isnan(y).any()),           'no nan allowed in x and y'
    assert (~(np.abs(np.diff(t,2)) > eps)).any(),                'x must contain equal time steps'
    assert isinstance(nbins, int) | (~nbins),                     'nbins must be integer'
    
    # get half size of samples
    n = t.size//2
    
    # get sample frequency 
    dt = t[1] - t[0]
    
    # generate x axis: frequencies
    X = np.fft.fftfreq(n*2, d=dt)[:n]
    
    # generate y axis: amplitudes
    Y = np.fft.fft(y)
    
    # use only abslute, real part of transformation and normalize to half frequencies
    Y = 2.*np.abs(Y[:n])/n

    # calculate logarithmic frequency binning
    if nbins:
        # transform to linear
        Xl, Yl = np.log10(X[1:]), np.log10(Y[1:])
        
        # generate bins
        bins = np.linspace(Xl[0], Xl[-1], nbins)
        
        # binning
        digi = np.digitize(Xl, bins)
        
        # get mean values for bins
        with wa.catch_warnings():
            wa.simplefilter("ignore")
            binm = np.array([Xl[digi == i].mean() for i in range(1, len(bins))])
            mv   = np.array([Yl[digi == i].mean() for i in range(1, len(bins))])
        
        # back-transform
        binm, mv = 10**binm, 10**mv
    
    if plot:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(2, 1)
        ax[0].plot(t,y)
        ax[0].set_xlabel('t')
        ax[0].set_ylabel('y')
        ax[1].plot(X, Y)
        if nbins:
            ax[1].plot(binm, mv, 'ro')
        ax[1].set_xlabel('Frequency [1/Unit of t]')
        ax[1].set_ylabel('Amplitude [Unit of y]')
        ax[1].set_xscale("log", nonposx='clip')
        ax[1].set_yscale("log", nonposx='clip')
        plt.show()
    
    if nbins:
        return X, Y, binm, mv
    else:
        return X, Y


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
from scipy import integrate as ig
from scipy.stats import beta
from pyjams.kernel_regression import kernel_regression, kernel_regression_h


__all__ = ['gap2lai', 'leafprojection']


def gap2lai(tgap, lgap, G, alpha, boot=False):
    """
        Computes effective leaf area index Le, total leaf area index Lt and
        clumping factor omega out of canopy gap fraction observations. When
        boot=False, for each gap fraction the output is calculated. If you give
        boot a number, the input is bootstrapped and mean +- standard error of
        the mean is returned as output. If you don't have large gap fraction
        lgap, you can give an array of zeros assuming random distribution of the
        leaves, Le equals Lt then.


        Definition
        ----------
        def gap2lai(tgap, lgap, G, alpha, boot=False):


        Input
        -----
        tgap         array(N), total gap fraction, e.g. fraction of the number
                     of all gap pixels in your cover photo divided by the total
                     number of pixels in your image. [-]
        lgap         array(N), large gap fraction, e.g. fraction of the number
                     of pixels only found in the largest gap of your cover photo
                     divided by the total number of pixels in your image. [-]
        G            float, value of the leaf projection function at view zenith
                     angle alpha [-]
        alpha        float, view zenith angle at which your observations are
                     made. 0 equals zenith, 90 equals horizont. [deg]


        Optional Input
        --------------
        boot         bool or int
                     if set False (default), for every gap fraction
                     value a separate output of Le, Lt and omega is computed.
                     If set to an int, the gap fractions are bootstrapped int times
                     and a mean +- standard error of the mean for Le, Lt and
                     omega is computed. WARNING: you are not allowed to average
                     the separate Le or Lt values after computation to get a
                     mean Le or Lt since you would introduce apparent clumping
                     correction.
                     If set True then boot=1.


        Output
        ------
        Lt           total leaf area index [m2leaf/m2ground]. if boot=False
                     (default), array(N) with Lt for each gap fraction. if
                     boot=int, array(2) with mean Lt and standard error of the
                     mean Lt
        Le           effective leaf area index [m2leaf/m2ground]. if boot=False
                     (default), array(N) with Le for each gap fraction. if
                     boot=int, array(2) with mean Le and standard error of the
                     mean Le
        omega        clumping index [-]. if boot=False (default), array(N) with
                     omega for each gap fraction. if boot=int, array(2) with
                     mean omega and standard error of the mean omega


        Literature
        ----------
        Macfarlane et al. (2007)
            Estimation of leaf area index in eucalypt forest using digital
            photography.
            Agricultural and Forest Meteorology 143, 176 - 188
        Ryu et al. (2010)
            On the correct estimation of effective leaf area index: Does it
            reveal information on clumping effects?
            Agricultural and Forest Meteorology 150, 463 - 472


        Examples
        --------
        >>> from autostring import astr

        >>> # Create some data
        >>> tgap  = np.arange(0.,1.1,0.1)
        >>> lgap  = np.arange(0.,1.1,0.1)*0.1
        >>> G     = 0.5
        >>> alpha = 53.7

        >>> # compute each observation separately
        >>> Lt, Le, omega = gap2lai(tgap, lgap, G, alpha)
        >>> print(astr(Lt, 2, join=True))
        --   2.81 1.97 1.47 1.11 0.84 0.62 0.43 0.27 0.13 0.00
        >>> print(astr(Le, 2, join=True))
        --   2.73 1.91 1.43 1.08 0.82 0.60 0.42 0.26 0.12 0.00
        >>> print('!', astr(omega, 2, join=True), '!')
        ! --   0.97 0.97 0.97 0.97 0.98 0.98 0.98 0.99 0.99 --   !

        >>> # compute mean and standard error of the mean
        >>> Lt, Le, omega = gap2lai(tgap, lgap, G, alpha, boot=10000)
        >>> print(astr(Lt, 1, join=True))
        0.9 0.2
        >>> print(astr(Le, 1, join=True))
        0.8 0.2
        >>> print(astr(omega, 1, join=True))
        1.0 0.0

        >>> # if you don't have lgap, Le and Lt are equal
        >>> Lt, Le, omega = gap2lai(tgap, np.zeros_like(tgap), G, alpha)
        >>> print(np.ma.all(Lt==Le))
        True


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Arndt Piayda

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Sep 2014
    """

    ###########################################################################
    # safety checks
    tgaps, lgaps = tgap.size, lgap.size
    assert (tgaps==lgaps) & (tgap.ndim==1) & (lgap.ndim==1), "gap2lai: tgap and lgap must be 1D and of same size"
    assert isinstance(G, float) & isinstance(alpha, float), "gap2lai: G and alpha must be of type float"
    assert (np.ma.max(tgap)<=1.) & (np.ma.max(lgap)<=1.), "gap2lai: max of lgap and tgap must be <=1."
    assert (np.ma.min(tgap)>=0.) & (np.ma.min(lgap)>=0.), "gap2lai: min of lgap and tgap must be >=0."
    assert np.ma.all((tgap-lgap)>=0.), "gap2lai: tgap must be larger or equal lgap"
    assert (np.signbit(G)==False) & (np.signbit(alpha)==False), "gap2lai: G and alpha must be positive"

    ###########################################################################
    # bootstrap input if desired
    if boot:
        tgap, lgap = gap2lai_bootstrap2(tgap, lgap, boot)
    #print(tgap)
    # radians from angle in degrees
    rad = np.deg2rad(alpha)
    # fraction of foliage cover, defined as proportion of ground area covered
    # by vertical projection of foliage and branches (crown cover)
    fc = 1.- lgap
    # proportion of ground area covered by vertical projection of foliage and
    # branches within the perimeter of the crowns of individual plants
    ff = 1.- tgap
    # crown porosity
    phi = 1. - ff/fc
    # effective leaf area index without clumping correction
    Le = -np.ma.log(1.-ff)*np.ma.cos(rad)/G
    Le = np.ma.where(Le==-0.0, 0.0, Le) # safety for case: tgap=1
    # clumping index
    omega = ((1.-phi)*np.ma.log(1.-ff))/(np.ma.log(phi)*ff)
    # total leaf area index with clumping correction
    Lt = -fc*np.ma.log(phi)*np.ma.cos(rad)/G
    # if bootstrap is not false, mean and standard error of the mean is returned
    if boot:
        Lt    = np.ma.array([np.ma.mean(Lt), np.ma.std(Lt)]).flatten()
        Le    = np.ma.array([np.ma.mean(Le), np.ma.std(Le)]).flatten()
        omega = np.ma.array([np.ma.mean(omega), np.ma.std(omega)]).flatten()

    return Lt, Le, omega


def leafprojection(alpha, theta, t360=False, kernel=False, min=0., max=90.,
                   step=5., h=None, boot=False):
    """
        Computes leaf projection funtion G for a certain view zenith angle alpha
        from measurements of leaf zenith angles theta (angle of the leaf surface
        normal to the zenith). When boot is set to an int, the input thetas are
        bootstrapped and mean +- standard error of the mean is returned as
        output. You can coose between the fast beta distribution function
        (default) or the more representative kernel regression.


        Definition
        ----------
        def leafprojection(alpha, theta, t360=False, kernel=False, min=0., max=90.,
                           step=5., h=None, boot=False):


        Input
        -----
        alpha        float, view zenith angle at which you want the leaf
                     projection value G. 0 equals zenith, 90 equals horizont.
                     [deg]
        theta        np.ma.array(N), leaf inclination zenith angle measurements
                     (angle of the leaf surface normal to the zenith) [deg] -180
                     to 180 with 0 pointing to zenith. If t360=True: leaf zenith
                     angle measurements [deg] 0 to 360 with 0 pointing to
                     zenith, clockwise.


        Optional Input
        --------------
        t360         bool, if False (default) theta  -180 to 180, else 0 to 360
        kernel       bool, if False (default) beta function is used, if True,
                     kernel regression is used (takes much more time) but closer
                     to measurements
        min          float, minimum of theta histogram (only used for kernel
                     regression, default: 0.)
        max          float, maximum of theta histogram (only used for kernel
                     regression, default: 90.)
        step         float, step size of theta histogram (only used for kernel
                     regression, default: 5.)
        h            float, if None (default) correlation lenght for kernel
                     regression is calculated from the input, else you can give
                     a float here.
        boot         bool or int, if set False (default), the original theta
                     distribution is used for the computations. If set to an
                     int, the theta distribution is bootstrapped int times and
                     a mean +- standard error of the mean for the projection
                     function G is computed


        Output
        ------
        G            float, value of the leaf projection function at the view
                     zenith angle alpha, given the leaf view zenith angle
                     distribution measurements theta [-]. If boot=int, tuple(2)
                     with mean G and standard error of the mean G at alpha


        Literature
        ----------
        Goel & Strebel (1984)
            Simple Beta Distribution Representation of Leaf Orientation in
            Vegetation Canopies.
            Agron. J., 1984, 76, 800-802
        Wang et al. (2007)
            Comparison of leaf angle distribution functions: Effects on
            extinction coefficient and fraction of sunlit foliage.
            Agricultural and Forest Meteorology, 2007, 143, 106 - 122


        Examples
        --------
        >>> # use beta ditribution for a view zenith angle of 57.3 deg and
        >>> # measured leaf inclination in -180 to 180 deg.
        >>> np.random.seed(1)
        >>> theta = np.ma.array(np.random.random(50)*180.)
        >>> print(np.ma.round(leafprojection(57.3, theta), 2))
        0.5

        >>> # use beta ditribution for a view zenith angle of 57.3 deg and
        >>> # measured leaf inclination in 0 to 360 deg.
        >>> np.random.seed(1)
        >>> theta = np.ma.array(np.random.random(50)*360.)
        >>> print(np.ma.round(leafprojection(57.3, theta, t360=True), 2))
        0.5

        >>> # use kernel regression for a view zenith angle of 57.3 deg and
        >>> # measured leaf inclination in 0 to 360 deg.
        >>> np.random.seed(1)
        >>> theta = np.ma.array(np.random.random(50)*360.)
        >>> print(np.ma.round(leafprojection(57.3, theta, t360=True, kernel=True), 2))
        0.5

        >>> # use kernel regression for a view zenith angle of 57.3 deg and
        >>> # measured leaf inclination in 0 to 360 deg. and set h fix to 1.0
        >>> np.random.seed(1)
        >>> theta = np.ma.array(np.random.random(50)*360.)
        >>> print(np.ma.round(leafprojection(57.3, theta, t360=True, kernel=True, h=1.0), 2))
        0.5

        >>> # use kernel regression for a view zenith angle of 57.3 deg and
        >>> # measured leaf inclination in 0 to 360 deg. Estimate the error with
        >>> # 10 bootstraps
        >>> np.random.seed(1)
        >>> theta = np.ma.array(np.random.random(50)*360.)
        >>> from autostring import astr
        >>> print(astr(leafprojection(57.3, theta, t360=True, kernel=True, boot=10), 3, joinall=True))
        0.503 0.003


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Arndt Piayda

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Sep 2014
    """

    assert not np.all(theta.mask) & np.ma.isMA(theta), 'leafprojection: theta is fully masked.'

    ###########################################################################
    # normalizing function for angles
    def t(thetal):
        return 2.*thetal/np.pi

    # beta distribution
    def f_beta(thetal, mu, nu):
        return beta.pdf(t(thetal), mu, nu)*2./np.pi

    # azimuth condition
    def vartheta(thetal, theta):
        return np.ma.arccos((1./np.ma.tan(theta)) * (1./np.ma.tan(thetal)))

    #undefined for 90 degrees due to tangens
    def phi(thetal, theta):
        return np.ma.where(np.ma.absolute((1./(np.ma.tan(theta)) * (1./np.ma.tan(thetal))))>1.,
                        np.ma.cos(theta)*np.ma.cos(thetal),
                        np.ma.cos(theta)*np.ma.cos(thetal)*(1. + (2./np.pi)*(np.ma.tan(
                        vartheta(thetal, theta))-vartheta(thetal, theta))))

    # thetal must be first argument, because first argument is taken for
    # integration by .quad
    def integ(thetal, theta, mu, nu):
        return phi(thetal, theta)*f_beta(thetal, mu, nu)
    def integ_k(thetal, theta, mbins, hist, h):
        ker = kernel_regression(mbins, hist, h, xout=np.array([t(thetal)]))
        return phi(thetal, theta)*ker*2./np.pi

    # leaf projection function with beta function
    def G(theta, mu, nu):
        if isinstance(theta, (int, float)):
            return ig.quad(integ, 0., np.pi/2., args=(theta, mu, nu))[0]
        else:
            out = np.array([ig.quad(integ, 0, np.pi/2.,args=(x, mu, nu))[0]
                             for x in theta])
            return np.ma.array(out, mask=theta.mask)

    # leaf projection function with kernel regression
    def G_ker(theta, mbins, hist, h):
        if isinstance(theta, (int, float)):
            return ig.quad(integ_k, 0, np.pi/2., args=(theta, mbins, hist, h))[0]
        else:
            out = np.array([ig.quad(integ_k, 0, np.pi/2.,args=(x, mbins, hist, h))[0]
                             for x in theta])
            return np.ma.array(out, mask=theta.mask)

    ###########################################################################
    # converting angles to 0-90 deg span and to rad
    if t360:
        theta = np.deg2rad(np.ma.where(theta<=90., theta,
                           np.ma.where(theta<=180., 180.-theta,
                           np.ma.where(theta<=270., theta-180., theta-270.))))
    else:
        theta = np.deg2rad(np.ma.where(np.ma.absolute(theta)<=90.,
                                       np.ma.absolute(theta),
                                       180.-np.ma.absolute(theta)))
    rad = np.deg2rad(alpha)

    if not kernel:
        #######################################################################
        # parameter determination for beta function
        if boot:
            theta = gap2lai_bootstrap(theta, boot)

        sig_0_2 = np.ma.mean(t(theta), axis=0)*(1.-np.ma.mean(t(theta), axis=0))
        sig_t_2 = np.ma.var(t(theta), axis=0)
        mu      = np.ma.mean(t(theta), axis=0)*(sig_0_2/sig_t_2-1.)
        nu      = (1.-np.ma.mean(t(theta), axis=0))*(sig_0_2/sig_t_2-1.)

        if boot:
            G = np.ma.array([G(rad, x[0], x[1]) for x in zip(mu, nu)])
            return np.ma.mean(G), np.ma.std(G)
        else:
            return G(rad, mu, nu)

    else:
        #######################################################################
        # parameter determination for kernel regression
        # histogram
        if boot:
            theta = gap2lai_bootstrap(theta, boot)

        min, max, step  = np.deg2rad(min), np.deg2rad(max), np.deg2rad(step)
        hist, bins, mbins = gap2lai_axishist(theta, min, max, step, t)

        # kernel h estimation
        if boot:
            G = np.ma.empty(boot)
            for i in range(boot):
                if not h:
                    h = kernel_regression_h(mbins[:,i], hist[:,i])
                G[i] = G_ker(rad, mbins[:,i], hist[:,i], h)
            return np.ma.mean(G), np.ma.std(G)
        else:
            h = kernel_regression_h(mbins, hist)
            return G_ker(rad, mbins, hist, h)

def gap2lai_axishist(x, min, max, step, t):
    """
    gap2lai: histogram along 0 axis of an array
    """
    inbins = np.arange(t(min),t(max+step),t(step))

    if x.ndim==1:
        hist, bins = np.histogram(t(x.compressed()), bins=inbins, density=True)
        mbins = bins[:-1]+t(step)/2.
    else:
        bs, xs = inbins.size, x.shape[1]
        hist, bins = np.ma.empty((bs-1, xs)), np.ma.empty((bs, xs))
        for i in range(xs):
            hist[:,i], bins[:,i] = np.histogram(t(x[:,i].compressed()), bins=inbins, density=True)
        mbins = bins[:-1,:]+t(step)/2.

    return hist, bins, mbins

def gap2lai_bootstrap(x, boot):
    """
    gap2lai: bootstrap a 1D array
    """
    xs = x.size
    assert x.ndim==1, "bootstrap: x and y must be 1D and of same size"
    assert isinstance(boot, int), "bootstrap: boot must be of type int"

    x_boot = np.ma.empty((xs, boot))
    for j in range(boot):
        ind = np.random.randint(xs,size=xs)
        x_boot[:,j] = x[ind]

    return x_boot

def gap2lai_bootstrap2(x, y, boot):
    """
    gap2lai: bootstrap two 1D arrays with same indices
    """
    xs, ys = x.size, y.size
    assert (xs==ys) & (x.ndim==1) & (y.ndim==1), "bootstrap: x and y must be 1D and of same size"
    assert isinstance(boot, int), "bootstrap: boot must be of type int"

    x_boot, y_boot = np.ma.empty(boot), np.ma.empty(boot)
    for j in range(boot):
        ind = np.random.randint(xs,size=xs)
        x_boot[j] = np.ma.mean(x[ind])
        y_boot[j] = np.ma.mean(y[ind])

    return x_boot, y_boot

if __name__ == '__main__':
     import doctest
     doctest.testmod()

     # from autostring import astr

     # # Create some data
     # tgap  = np.arange(0.,1.1,0.1)
     # lgap  = np.arange(0.,1.1,0.1)*0.1
     # G     = 0.5
     # alpha = 53.7

     # # compute each observation separately
     # Lt, Le, omega = gap2lai(tgap, lgap, G, alpha)
     # print(np.ma.round(Lt,2))
     # print(astr(Lt, 2, join=True))
     # # --   2.81 1.97 1.47 1.11 0.84 0.62 0.43 0.27 0.13 0.00
     # print(np.ma.round(Le,2))
     # print(astr(Le, 2, join=True))
     # # --   2.73 1.91 1.43 1.08 0.82 0.6 0.42 0.26 0.12 0.00
     # print(np.ma.round(omega,2))
     # print(astr(omega, 2, join=True))
     # # --   0.97 0.97 0.97 0.97 0.98 0.98 0.98 0.99 0.99 --

     # # compute mean and standard error of the mean
     # Lt, Le, omega = gap2lai(tgap, lgap, G, alpha, boot=10000)
     # print(np.ma.round(Lt,1))
     # print(astr(Lt, 1, join=True))
     # # 0.9 0.2
     # print(np.ma.round(Le,1))
     # print(astr(Le, 1, join=True))
     # # 0.8 0.2
     # print(np.ma.round(omega,1))
     # print(astr(omega, 1, join=True))
     # # 1.0 0.0
     # # if you don't have lgap, Le and Lt are equal
     # Lt, Le, omega = gap2lai(tgap, np.zeros_like(tgap), G, alpha)
     # print(np.ma.all(Lt==Le))
     # # True



#!/usr/bin/env python
import numpy as np
"""
    Defines signatures of discharge time series.

    They are:
        autocorrelation
        flow duration curves
        rising and declining limb densities
        maximum monthly flow
        moments
        peak distribution

    Translation of the module mo_mrm_signatures.f90 by mainly Juliane Mai, Jun 2015
    RunoffRatio and ZeroFlowRatio were not ported.


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014-2016 Matthias Cuntz, Remko Nijzink, Juliane Mai - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written, MC, May 2016
"""

all = ['autocorrelation', 'flowdurationcurve', 'limbdensities', 'maximummonthlyflow', 'moments', 'peakdistribution']

# --------------------------------------------------------------------

def autocorrelation(dat, lags):
    """
        NAME
        autocorrelation

        PURPOSE
        Autocorrelation of a given data series.

        Calculates the autocorrelation of a data series at given lags.
        The function is basically a wrapper of the function jams.correlate.

        DEFINITION
        def autocorrelation(dat, lags):

        INTENT(IN)
        data   1D array_like of data
        lags   scalar or 1D array_like of lags where autocorrelation is requested
               positive and negative lags are possible

        RETURN
        autocorrelation of data at given lags

        RESTRICTIONS
        Works only with 1D arrays

        EXAMPLE
            None

        LITERATURE
            Used as hydrologic signature with lag 1 in
               Euser, T., Winsemius, H. C., Hrachowitz, M., Fenicia, F., Uhlenbrook, S., & Savenije, H. H. G. (2013).
               A framework to assess the realism of model structures using hydrological signatures.
               Hydrology and Earth System Sciences, 17(5), 1893-1912. doi:10.5194/hess-17-1893-2013

        HISTORY
        Written, Juliane Mai, Jun 2015 in Fortran
        Modified, Matthias Cuntz, May 2016 - Python port
    """
    from pyjams.jams.correlate import correlate
    autocorrelation = correlate(dat, dat)
    return autocorrelation[lags]

# --------------------------------------------------------------------

def flowdurationcurve(dat, quantiles=None, concavity_index=False,
                      mid_segment_slope=False, mhigh_segment_volume=False,
                      high_segment_volume=False, low_segment_volume=False):
    """
        NAME
            flowdurationcurve

        PURPOSE
        Flow duration curves.

        Calculates the flow duration curves for a given data vector. The Flow duration curve at a
        certain quantile x is the data point p where x% of the data points are above the value p.
        Hence the function percentile of the module mo_percentile is used. But percentile is
        determining the point p where x% of the data points are below that value. Therfore, the
        given quantiles are transformed by (1.0-quantile) to get the percentiles of exceedance probabilities.

        Optionally, the concavity index CI can be calculated [Zhang2014]. CI is defined by
                \f[ CI = \frac{q_{10\%}-q_{99\%}}{q_{1\%}-q_{99\%}} \f]
        where \f$ q_{x} \f$ is the data point where x% of the data points are above that value.
        Hence, exceedance probabilities are used.\n

        Optionally, the FDC mid-segment slope \f$FDC_{MSS}\f$ as used by Shafii et. al (2014) can be returned.
        The \f$FDC_{MSS}\f$ is defined as
                \f[ FDC_{MSS} = \log(q_{m_1})-\log(q_{m_2}) \f]
        where \f$ m_1 \f$ and \f$ m_2 \f$ are the lowest and highest flow exceedance probabilities within the
        midsegment of FDC. The settings \f$m_1=0.2\f$ and \f$0.7\f$ are used by Shafii et. al (2014) and are
        implemented like that.\n

        Optionally, the FDC medium high-segment volume \f$FDC_{MHSV}\f$ as used by Shafii et. al (2014) can be
        returned. The \f$FDC_{MHSV}\f$ is defined as
                \f[ FDC_{MHSV} = \sum_{h=1}^{H} q_h \f]
        where \f$ h=1,2,...,H \f$ are flow indeces located within the high-flow segment (exceedance probabilities
        lower than \f$m_1\f$). \f$H\f$ is the index of the maximum flow. The settings \f$m_1=0.2\f$ is used here
        to be consistent with the definitions of the low-segment (0.7-1.0) and the mid-segment (0.2-0.7).\n

        Optionally, the FDC high-segment volume \f$FDC_{HSV}\f$ as used by Shafii et. al (2014) can be returned.
        The \f$FDC_{HSV}\f$ is defined as
                \f[ FDC_{HSV} = \sum_{h=1}^{H} q_h \f]
        where \f$ h=1,2,...,H \f$ are flow indeces located within the high-flow segment (exceedance probabilities
        lower than \f$m_1\f$). \f$H\f$ is the index of the maximum flow. The settings \f$m_1=0.02\f$ is used by
        Shafii et. al (2014) and is implemented like that.\n

        Optionally, the FDC low-segment volume \f$FDC_{LSV}\f$ as used by Shafii et. al (2014) can be returned.
        The \f$FDC_{LSV}\f$ is defined as
                \f[ FDC_{LSV} = -\sum_{l=1}^{L} (\log(q_l) - \log(q_L)) \f]
        where \f$ l=1,2,...,L \f$ are flow indeces located within the low-flow segment (exceedance probabilities
        larger than \f$m_1\f$). \f$L\f$ is the index of the minimum flow. The settings \f$m_1=0.7\f$ is used by
        Shafii et. al (2014) and is implemented like that.\n

        CALLING SEQUENCE
        def flowdurationcurve(dat, quantiles=None, concavity_index=False,
                              mid_segment_slope=False, mhigh_segment_volume=False,
                              high_segment_volume=False, low_segment_volume=False):

        INTENT(IN)
        data              1D array_like data series

        INTENT(IN), OPTIONAL
        quantiles              Scalar or 1D array_like percentages of exceedance
        concavity_index        True: concavity index as defined by Sauquet et al. (2011)
        mid_segment_slope      True: mid-segment slope as defined by Shafii et al. (2014)
        mhigh_segment_volume   True: medium high-segment volume
        high_segment_volume    True: high-segment volume as defined by Shafii et al. (2014)
        low_segment_volume     True: low-segment volume as defined by Shafii et al. (2014)

        RETURN
        Flow duration curve at resp. quantiles or optional signatures of the flow duration curve

        RESTRICTIONS
        Thresholds in mid_segment_slope, mhigh_segment_volume, high_segment_volume, low_segment_volume are hard coded.

        EXAMPLE
            None

        LITERATURE
            FDC is used as hydrologic signature (quantiles not specified) in
               Euser, T., Winsemius, H. C., Hrachowitz, M., Fenicia, F., Uhlenbrook, S., & Savenije, H. H. G. (2013).
               A framework to assess the realism of model structures using hydrological signatures.
               Hydrology and Earth System Sciences, 17(5), 1893-1912. doi:10.5194/hess-17-1893-2013
            Concavity Index used as hydrologic signature in
               Zhang, Y., Vaze, J., Chiew, F. H. S., Teng, J., & Li, M. (2014).
               Predicting hydrological signatures in ungauged catchments using spatial interpolation, index model, and
               rainfall-runoff modelling.
               Journal of Hydrology, 517(C), 936-948. doi:10.1016/j.jhydrol.2014.06.032
            Concavity index is defined using exceedance probabilities by
               Sauquet, E., & Catalogne, C. (2011).
               Comparison of catchment grouping methods for flow duration curve estimation at ungauged sites in France.
               Hydrology and Earth System Sciences, 15(8), 2421-2435. doi:10.5194/hess-15-2421-2011
            mid_segment_slope, high_segment_volume, low_segment_volume used as hydrologic signature in
               Shafii, M., & Tolson, B. A. (2015).
               Optimizing hydrological consistency by incorporating hydrological signatures into model calibration objectives.
               Water Resources Research, 51(5), 3796-3814. doi:10.1002/2014WR016520

        HISTORY
        Written Remko Nijzink and Juliane Mai, March 2014 - June 2015 in Fortran
        Modified, Matthias Cuntz, May 2016 - Python port
    """
    if isinstance(dat, np.ma.MaskedArray):
        mdat = dat.data[~dat.mask]
    else:
        mdat = dat

    allind = concavity_index+mid_segment_slope+mhigh_segment_volume+high_segment_volume+low_segment_volume
    if quantiles is not None:
        if allind > 0:
            raise IOError('Either quantiles or indexes can be returned.')
    else:
        if allind > 1:
            raise IOError('Only one index can be returned.')

    if quantiles is not None:
        fdc = np.percentile(mdat, (1.-np.array(quantiles))*100.)
        return fdc
    elif concavity_index:
        p10 = np.percentile(mdat, (1.-0.10)*100.)
        p99 = np.percentile(mdat, (1.-0.99)*100.)
        p01 = np.percentile(mdat, (1.-0.01)*100.)
        concavity_index = (p10-p99) / (p01-p99)
        return concavity_index
    elif mid_segment_slope:
        p20 = np.percentile(mdat, (1.-0.20)*100.)
        p70 = np.percentile(mdat, (1.-0.70)*100.)
        mid_segment_slope = np.log(p20) - np.log(p70)
        return mid_segment_slope
    elif mhigh_segment_volume:
        # medium high-flows are defined to be between 0.0 and 0.2 as to be constistent
        # with the mid-segment (0.2-0.7) and low-segment (0.7-1.0) definitions
        p20 = np.percentile(mdat, (1.-0.20)*100.)
        ii = np.where(mdat>=p20)[0]
        mhigh_segment_volume = np.sum(mdat[ii])
        return mhigh_segment_volume
    elif high_segment_volume:
        # high-flows are defined to be between 0.0 and 0.02 by Shafii et. al (2014)
        p02 = np.percentile(mdat, (1.-0.02)*100.)
        ii = np.where(mdat>=p02)[0]
        high_segment_volume = np.sum(mdat[ii])
        return high_segment_volume
    elif low_segment_volume:
        # low-flows are defined to be between 0.7 and 1.0 by Shafii et. al (2014)
        min_flow_value = np.amin(mdat)
        p70 = np.percentile(mdat, (1.-0.7)*100.)
        ii = np.where(mdat<=p70)[0]
        low_segment_volume = -1.0 * np.sum(np.log(mdat[ii]) - np.log(min_flow_value))
        return low_segment_volume
    else:
        raise ValueError('Should not be here.')

# --------------------------------------------------------------------

def limbdensities(dat):
    """
        NAME
        limbdensities

        PURPOSE
        Calculates limb densities

        Calculates rising and declinging limb densities. The peaks of the given series are
        first determined by looking for points where preceding and subsequent datapoint are lower.
        Second, the number of datapoints with rising values (nrise) and declining values (ndecline)
        are counted basically by comparing neighbors.

        The duration the data increase (nrise) divided by the number of peaks (npeaks)
        gives the rising limb density RLD
            \f[ RLD=t_{rise}/n_{peak} \f]
        whereas the duration the data decrease (ndecline) divided by the number of peaks (npeaks)
        gives the declining limb density DLD
            \f[ DLD=t_{fall}/n_{peak}. \f]

        CALLING SEQUENCE
        def limbdensities(dat, rld=False, dld=False):

        INTENT(IN)
        dat    1D array_like data series

        RETURN
        rising limb density, declining limb density

        RESTRICTIONS
            None

        EXAMPLE
            None

        LITERATURE
            Rising limb density used as hydrologic signature in
               Euser, T., Winsemius, H. C., Hrachowitz, M., Fenicia, F., Uhlenbrook, S., & Savenije, H. H. G. (2013).
               A framework to assess the realism of model structures using hydrological signatures.
               Hydrology and Earth System Sciences, 17(5), 1893-1912. doi:10.5194/hess-17-1893-2013

        HISTORY
            Written Remko Nijzink and Juliane Mai, March 2014 - Jun 2015 in Fortran
            Modified Matthias Cuntz, May 2016 - Python port
    """
    if isinstance(dat, np.ma.MaskedArray):
        ndat = dat.count()
    else:
        ndat = dat.size

    thres_rise = 1.0
    dd = np.ma.diff(dat)
    nrise = np.sum((dd-thres_rise) > 0.)
    ndecline = ndat - nrise
    # dd[0:-1]*dd[1:]<0 gives min and max: multiply with dd[0:-1] -> <0 are maxima
    dd1 = dd[0:-1]*dd[1:]
    dextreme = np.where(dd1 < 0., dd1, 0.)
    npeak = np.sum(dextreme*dd[0:-1] < 0.)

    if npeak>0:
        return float(nrise)/float(npeak), float(ndecline)/float(npeak)
    else:
        return 0., 0.

# --------------------------------------------------------------------

def maximummonthlyflow(date, dat):
    """
    NAME
    maximummonthlyflow

    PURPOSE
    Maximum of average flows per months.

    Maximum of average flow per month is defined as
       \f[ max_{monthly flow} = Max( F(i), i=1,..12 ) \f]
    where \$f F(i) $\f is the average flow of month i.

    CALLING SEQUENCE
    def maximummonthlyflow(date, dat):

    INTENT(IN)
    date          1d array_like julian date
    data          1d array_like of data

    RETURN
    Maximum of average flow per month

    RESTRICTIONS
    None

    EXAMPLE
        None

    LITERATURE
        used as hydrologic signature in
           Shafii, M., & Tolson, B. A. (2015).
           Optimizing hydrological consistency by incorporating hydrological signatures into model calibration objectives.
           Water Resources Research, 51(5), 3796-3814. doi:10.1002/2014WR016520

    HISTORY
    Written Juliane Mai, Jun 2015 in Fortran
    Modified Matthias Cuntz, May 2016 - Python port
    """
    from pyjams.means import means
    mdate, mdat = means(date, dat, month=True)
    return mdat.max()

# --------------------------------------------------------------------

def moments(dat, mean_data=False, stddev_data=False, median_data=False,
            max_data=False, mean_log=False, stddev_log=False, median_log=False, max_log=False):
    """
    NAME
    moments

    PURPOSE
    Moments of data and log-transformed data, e.g. mean and standard deviation.

    Returns several moments of data series given, i.e.
        * mean               of data
        * standard deviation of data
        * median             of data
        * maximum/ peak      of data
        * mean               of log-transformed data
        * standard deviation of log-transformed data
        * median             of log-transformed data
        * maximum/ peak      of log-transformed data

    CALLING SEQUENCE
    def moments(dat, mean_data=False, stddev_data=False, median_data=False,
                max_data=False, mean_log=False, stddev_log=False, median_log=False, max_log=False):

    INTENT(IN)
    dat        1D array_like of data

    INTENT(OUT), OPTIONAL
    mean_data     mean               of data
    stddev_data   standard deviation of data
    median_data   median             of data
    max_data      maximum/ peak      of data
    mean_log      mean               of log-transformed data
    stddev_log    standard deviation of log-transformed data
    median_log    median             of log-transformed data
    max_log       maximum/ peak      of log-transformed data

    RETURN
    moment(s) of data depending on optional arguments
    order is given by the above order of the parameters

    RESTRICTIONS
    None

    EXAMPLE
    None

    LITERATURE
        mean_log and stddev_log used as hydrologic signature in
           Zhang, Y., Vaze, J., Chiew, F. H. S., Teng, J., & Li, M. (2014).
           Predicting hydrological signatures in ungauged catchments using spatial interpolation, index model, and
           rainfall-runoff modelling.
           Journal of Hydrology, 517(C), 936-948. doi:10.1016/j.jhydrol.2014.06.032
        mean_data, stddev_data, median_data, max_data, mean_log, and stddev_log used as hydrologic signature in
           Shafii, M., & Tolson, B. A. (2015).
           Optimizing hydrological consistency by incorporating hydrological signatures into model calibration objectives.
           Water Resources Research, 51(5), 3796-3814. doi:10.1002/2014WR016520

    HISTORY
    Written Juliane Mai, Jun 2015 in Fortran
    Modified Matthias Cuntz, May 2016 - Python port
    """
    allind = mean_data+stddev_data+median_data+max_data+mean_log+stddev_log+median_log+max_log
    if allind == 0:
        raise IOError('At least one moment must be returned.')

    if isinstance(dat, np.ma.MaskedArray):
        mdat = dat.data[~dat.mask]
    else:
        mdat = dat

    out = []
    if mean_data:   out.append(np.mean(mdat))
    if stddev_data: out.append(np.std(mdat, ddof=1))
    if median_data: out.append(np.median(mdat))
    if max_data:    out.append(np.amax(mdat))

    allind = mean_log+stddev_log+median_log+max_log
    if allind > 0:
        undef = -9999
        ii = np.where(mdat > 0.)[0]
        if ii.size > 0:
            logdat = np.log(mdat[ii])
            if mean_log:   out.append(np.mean(logdat))
            if stddev_log: out.append(np.std(logdat, ddof=1))
            if median_log: out.append(np.median(logdat))
            if max_log:    out.append(np.amax(logdat))
        else:
            if mean_log:   out.append(undef)
            if stddev_log: out.append(undef)
            if median_log: out.append(undef)
            if max_log:    out.append(undef)

    return out

# --------------------------------------------------------------------

def peakdistribution(dat, quantiles=None, slope=False):
    """
    NAME
    peakdistribution

    PURPOSE
    Calculates the peak distribution.

    First, the peaks of the time series given are identified. For the peak distribution
    only this subset of data points are considered. Second, the peak distribution at the
    quantiles given is calculated. Calculates the peak distribution at the quantiles given
    using mo_percentile. Since the exceedance probabilities are usually used in
    hydrology the function percentile is used with (1.0-quantiles). \n

    Optionally, the slope of the peak distribution between 10th and 50th percentile, i.e.
       \f[ slope = \frac{\mathrm{peak\_{data}}_{0.1}-\mathrm{peak\_{data}}_{0.5}}{0.9-0.5} \f]
    can be returned.\n

    CALLING SEQUENCE
    def peakdistribution(dat, quantiles=None, slope=False):

    INTENT(IN)
    dat                       1D array_like data array

    INTENT(IN), OPTIONAL
    quantiles                 requested quantiles for distribution
    slope                     True: slope of the peak distribution between 10th and 50th percentile

    RETURN
    Distribution of peak values at resp. quantiles   or   slope of the peak distribution

    RESTRICTIONS
    None

    EXAMPLE
    None

    LITERATURE
        slope used as hydrologic signature in
           Euser, T., Winsemius, H. C., Hrachowitz, M., Fenicia, F., Uhlenbrook, S., & Savenije, H. H. G. (2013).
           A framework to assess the realism of model structures using hydrological signatures.
           Hydrology and Earth System Sciences, 17(5), 1893-1912. doi:10.5194/hess-17-1893-2013

    HISTORY
    Written Remko Nijzink and Juliane Mai, March 2014 - Jun 2015 in Fortran
    Modified Matthias Cuntz, May 2016 - Python port
    """
    if quantiles is not None:
        if slope:
            raise IOError('Either quantiles or slope of peak distribution can be returned.')
    else:
        if not slope:
            raise IOError('Either quantiles or slope of peak distribution must be returned.')

    # find peaks
    data_peak = []
    for i in range(1,len(dat)-1):
        if ((dat[i-1] <= dat[i]) and (dat[i+1] <= dat[i])): data_peak.append(dat[i])
    data_peak = np.array(data_peak)

    if quantiles is not None:
        return np.percentile(data_peak, (1.-np.array(quantiles))*100.)
    else:
        # calculate slope between 10% and 50% quantiles, per definition
        quantiles = [0.1, 0.5]
        p10, p50 = np.percentile(data_peak, (1.-np.array(quantiles))*100.)
        return (p10-p50)/(0.9-0.5)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
"""
This is the unittest for romanliterals module.

python -m unittest -v tests/test_romanliterals.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_romanliterals.py

"""
import unittest


class TestRomanliterals(unittest.TestCase):
    """
    Tests for romanliterals.py
    """

    def test_int2roman(self):
        from pyjams import int2roman

        assert int2roman(1) == 'I'
        assert int2roman(19) == 'XIX'
        assert int2roman(159) == 'CLIX'
        assert int2roman(159, lower=True) == 'clix'

    def test_roman2int(self):
        from pyjams import roman2int

        assert roman2int('I') == 1
        assert roman2int('i') == 1
        assert roman2int('iv') == 4
        assert roman2int('MCCCLIV') == 1354


if __name__ == "__main__":
    unittest.main()
